@@ -3,7 +3,6 @@ import cv2
 from tqdm.auto import tqdm
 import matplotlib.pyplot as plt
 import numpy as np
-from tensorflow.keras.applications.resnet50 import ResNet50, preprocess_input
 from sklearn.decomposition import PCA
 import plotly.express as px
 import pandas as pd
@@ -15,6 +14,7 @@ import yaml
 from ipywidgets import HTML, VBox
 from plotly import graph_objects as go
 import plotly.io as pio
+from sklearn.metrics.pairwise import cosine_similarity
 
 pio.renderers.default = "notebook"
 
@@ -36,15 +36,16 @@ class Owl:
     This class represents image dataset from directory (we call it gallery). Currently it supports methods 
     for plotting image data and tools like apply or undersampling.  
     """
-    def __init__(self, gallery):
+    def __init__(self, gallery, im_size):
         """
         Initializes dataset from directory
         Parameters:
         - gallery: path to directory with images
+        - im_size: tuple, size of images after loading
         """
         self.gallery = gallery
         self.filenames = os.listdir(self.gallery)
-        self.data = np.array([load_img(f"{self.gallery}/{image_file}", (420,420)) for image_file in tqdm(self.filenames, desc="Loading")])
+        self.data = np.array([load_img(f"{self.gallery}/{image_file}", im_size) for image_file in tqdm(self.filenames, desc="Loading")])
         self.cached_features_dataframe = None
         
 
@@ -91,18 +92,20 @@ class Owl:
         plt.show()
         
         
-    def _get_features(self):
+    def _get_features(self, model, preprocess_function):
         """
         Extracts features from images (self.data) with Resnet50 and PCA. After 
         processing returns pd.DataFrame with components x & y and filenames.
         Because feature extraction is long operation we are caching data 
         in self.cached_features_dataframe.
+        Parameters:
+        - model: keras application (nn) to feature extraction
+        - preprocess_function: function for input preprocessing before forward pass
         """
         if self.cached_features_dataframe is not None:
             return self.cached_features_dataframe
         else:
-            model = ResNet50(weights='imagenet', include_top=False)
-            images = self.apply(preprocess_input)
+            images = self.apply(preprocess_function)
             features = model.predict(images)
             features = features.reshape(len(features), np.prod(features.shape[1:]))
             
@@ -118,10 +121,12 @@ class Owl:
             df["filenames"] = self.filenames
             df["filenames"] = df["filenames"].apply(lambda x: f"{self.gallery}/{x}")
             self.cached_features_dataframe = df
+            self.cached_pca_features = pca_features
             
             return df
         
-    def distplot(self, kind="scatter"):
+        
+    def distplot(self, feature_params, kind="scatter"):
         """
         Creates 2d plot based on extracted features. You can use to show distribution
         of images on your dataset and analyze clusters.
@@ -130,8 +135,12 @@ class Owl:
             * scatter - plotly scatter and filename on hover
             * iscatter - interactive scatter that show image on hover inside jupyter noteboook.
             * joint - hex 2dim plot (sns.jointplot used)
+        - feature_params: dictionary of {
+            model: tf.keras.application,
+            preprocess_function: tf.keras.application preprocess_input function
+        }
         """
-        df = self._get_features()
+        df = self._get_features(feature_params["model"], feature_params["preprocess_function"])
         if kind == "scatter":
             fig = px.scatter(df, x="x", y="y", title="Images Distribution", hover_data="filenames")
             fig.show("notebook")
@@ -149,20 +158,24 @@ class Owl:
                 html.value = template.format(**row)
 
             fig = go.FigureWidget(data=fig.data, layout=fig.layout)
-            fig.data[0].on_hover(update)
+            fig.data[0].on_click(update)
 
             return VBox([fig, html])
             
             
-    def undersample(self, target_size):
+    def undersample(self, target_size, feature_params):
         """
         Undersamples dataset with kde-method. Returns filepaths of undersampled set and plots main sample
         and undersampled set on top to view distribution.
         Parameters:
         - target_size: target size of subsample, must be lower than size of dataset.
+        - feature_params: dictionary of {
+            model: tf.keras.application,
+            preprocess_function: tf.keras.application preprocess_input function
+        }
         """
         assert target_size < len(self.data)
-        df = self._get_features()
+        df = self._get_features(feature_params["model"], feature_params["preprocess_function"])
         data = np.array(list(zip(df["x"].values,df["y"].values)))
         kde = scipy.stats.gaussian_kde(data.T)
         p = 1 / kde.pdf(data.T)
@@ -177,6 +190,33 @@ class Owl:
         
         return df.filenames.values[idx]
     
+    
+    def search_outliers(self, threshold, copy_to=None):
+        """
+        Search outliers in image dataset. Counts cosine similarity btw images. Then 
+        takes samples with lowest scores based on threshold. All possible outliers will 
+        be copied to folder provided in copy_to parameter.
+        Parameters:
+        - threshold: value in (0, 1) interval to select best matches.
+        - copy_to: directory, where you want to copy images. if folder not exists we'll
+        create it.
+        """
+        matrix = cosine_similarity(self.apply(lambda x: x.flatten()))
+        scores = {}
+        
+        for row, path in zip(matrix, self.filenames):
+            scores[path] = sum(row)
+        
+        scores = dict(sorted(scores.items(), key=lambda x: x[1]))
+        bound = int(len(scores) * threshold)
+        files = list(scores.keys())[:bound]
+        
+        if copy_to is None:
+            return files
+        else:
+            self.copy_subsample_to(files, copy_to)
+
+    
     def copy_to(self, destination):
         """
         Copies all images from gallery to another gallery called destination.
@@ -189,6 +229,28 @@ class Owl:
         for file in tqdm(self.filenames, desc="Copying"):
             shutil.copy(os.path.join(self.gallery, file), destination)
             
+        
+    def copy_subsample_to(self, files, destination):
+        """
+        Copies all images with given filenamesfrom gallery to another gallery called destination.
+        Parameters:
+        - files: which files to copy
+        - destination: directory, where you want to copy images. if folder not exists we'll
+        create it.
+        """
+        if not os.path.exists(destination):
+            os.makedirs(destination)
+        for file in tqdm(files, desc="Copying"):
+            shutil.copy(os.path.join(self.gallery, file), destination)
+            
+            
+    def as_iterable(self):
+        """
+        Returns zipped object for loop over image and path.
+        """
+        return zip(self.data, self.filenames)
+            
+            
             
 class DataChecker:
     """
@@ -222,6 +284,7 @@ class DataChecker:
             contents = yaml.safe_load(metafile)
         return contents
     
+    
     def _parse_detection_annotation(self, filepath):
         """
         Parse annotation file in detection task. Returns list of floats (coordinates) 
@@ -232,6 +295,7 @@ class DataChecker:
             contents = list(map(float, contents))
             contents[0] = int(contents[0])
         return contents
+    
       
     def _parse_classification_annotation(self, filepath):
         """
@@ -242,6 +306,7 @@ class DataChecker:
             contents = int(contents[0])
         return contents
     
+    
     def check(self):
         """
         Checks dataset validity based on self.task. Finally generates report 
